#include "gps.h"
#include <string.h>
#include "driver/uart.h"
#include "nmea.h"
#include "ubx.h"

#define GPS_TX_PIN 17
#define GPS_RX_PIN 16
#define GPS_UART UART_NUM_1
#define GPS_UART_BUFFER_SIZE 4096
#define MAX_LINE_SIZE 255

#define DEFAULT_BAUD_RATE 9600
//#define HIGHSPEED_BAUD_RATE 19200
//#define HIGHSPEED_BAUD_RATE 38400
//#define HIGHSPEED_BAUD_RATE 57600
#define HIGHSPEED_BAUD_RATE 115200

#define STR1(x)  #x
#define STR(x)  STR1(x)

// default port for broadcasting
#define BROADCAST_PORT 10110

static void initUART(uart_port_t uart, int speed) {
    uart_config_t myUartConfig;
    myUartConfig.baud_rate           = speed;
    myUartConfig.data_bits           = UART_DATA_8_BITS;
    myUartConfig.parity              = UART_PARITY_DISABLE;
    myUartConfig.stop_bits           = UART_STOP_BITS_1;
    myUartConfig.flow_ctrl           = UART_HW_FLOWCTRL_DISABLE;
    myUartConfig.rx_flow_ctrl_thresh = 122;

    uart_param_config(uart, &myUartConfig);

    uart_set_pin(uart,
                GPS_RX_PIN,         // TX
                GPS_TX_PIN,         // RX
                UART_PIN_NO_CHANGE, // RTS
                UART_PIN_NO_CHANGE  // CTS
    );

    uart_driver_install(uart, GPS_UART_BUFFER_SIZE, GPS_UART_BUFFER_SIZE, 0, NULL, 0);
}

void deinitUART(uart_port_t uart) {
    uart_driver_delete(uart);
}

static void readLine(uart_port_t uart, uint8_t *line) 
{
    int size;
    uint8_t *ptr = line;
    while(ptr-line < MAX_LINE_SIZE) {
        size = uart_read_bytes(uart, ptr, 1, portMAX_DELAY);
        if (size == 1) {
            if (*ptr == '\n') {
                ptr++;
                *ptr = 0;
                break;
            }
            ptr++;
        }
    }
}

void readAll(uart_port_t uart)
{
    uint8_t buf[1024];
    while (1) {
        size_t available = 0;
        if (ESP_OK != uart_get_buffered_data_len(uart, &available))
            return;
        if (! available)
            return;
        uart_read_bytes(uart, buf, available > sizeof(buf) ? sizeof(buf) : available, 0);
    }
}

// read UBX protocol message
// returns zero on success
static int readUbxReply(uart_port_t uart, uint8_t *buf, int maxSize, int *msgStart, int *msgLength) 
{
    int readed = 0;

    while (readed < maxSize) {
        int sz = uart_read_bytes(uart, buf + readed, 1, portMAX_DELAY);
        if (sz) {
            readed += sz;
            if (ubx_is_message_reeived(buf, readed, msgStart, msgLength))
                return 0;
        }
    }
    return -1;
}


static void setPortSpeed(uart_port_t uart)
{
    char buf[80];
    
    initUART(uart, DEFAULT_BAUD_RATE);
    nmea_make_message(buf, "$PUBX,41,1,0003,0003," STR(HIGHSPEED_BAUD_RATE) ",0");
    uart_write_bytes(uart, buf, strlen(buf));
    uart_wait_tx_done(uart, 1000);
    uart_set_baudrate(uart, HIGHSPEED_BAUD_RATE);
    //deinitUART(uart);
}


static int sendUbxMessage(uart_port_t uart, uint8_t *msg, int size)
{
    uint8_t buf[1024];
    int replyOffset, replyLength;

    readAll(uart);
    uart_write_bytes(uart, (char*)msg, size);
    if (! readUbxReply(uart, buf, sizeof(buf), &replyOffset, &replyLength)) {
        if ((0x05 == buf[replyOffset + 2]) && (0x01 == buf[replyOffset + 3])) {
            printf("ACK\n");
            return 0;
        } else if ((0x05 == buf[replyOffset + 2]) && (0x00 == buf[replyOffset + 3])) {
            printf("NACK\n");
        }
    } else {
        printf("ERROR: can't get UBX reply\n");
    }
    return 1;
}

// set GPS update frequency
// rate in herz
static void configRate(uart_port_t uart, int rate)
{
    uint8_t buf[32];
    int size;

    size = ubx_make_rate_message(buf, rate);
    sendUbxMessage(uart, buf, size);
    size = ubx_make_rates_message(buf, rate);
    sendUbxMessage(uart, buf, size);
    size = ubx_make_platform_model_message(buf, 5);
    sendUbxMessage(uart, buf, size);
}


static bool startsWith(uint8_t *line, char *start)
{
    for (int i = 0; start[i]; i++) {
        if (line[i] != (uint8_t)start[i])
            return false;
    }
    return true;
}


void gps_task(void *param)
{
    uint8_t line[MAX_LINE_SIZE+1];
    
    setPortSpeed(GPS_UART);
    //initUART(GPS_UART, HIGHSPEED_BAUD_RATE);
    configRate(GPS_UART, 5);
//    tcp_server_init(BROADCAST_PORT);
    while (1) {
        readLine(GPS_UART, line);
        if (! nmea_is_checksum_good((char*)line))
            printf("Invalid checksum in line %s", line);
        else {
            //printf("%s", line);
            if (line[0] && (startsWith(line, "$GNGLL") || startsWith(line, "$GNVTG")/* || 
                    startsWith(line, "$GPGLL") || startsWith(line, "$GPVTG")*/))
                printf("%s", line);
//                tcp_server_send((char*)line);
                /*if (bcast_send(BROADCAST_PORT, (char*)line)) {
                    printf("Error sending broadcat message\n");
                }*/
        }
//        tcp_server_update();
    }
//    tcp_server_done();
    deinitUART(GPS_UART);
}

